#!/usr/bin/env python3

"""
(C) 2022,2023,2025 Jack Lloyd

Botan is released under the Simplified BSD License (see license.txt)
"""

import datetime
import hashlib
import json
import multiprocessing
from multiprocessing.pool import ThreadPool
import optparse # pylint: disable=deprecated-module
import os
import random
import re
import shlex
import sqlite3
import subprocess
import sys
import time
import uuid

enabled_checks = [
    'bugprone-*',
    'cert-*',
    'clang-analyzer-*',
    'cppcoreguidelines-*',
    'hicpp-*',
    'misc-*',
    'modernize-*',
    'performance-*',
    'portability-*',
    'readability-*',
]

# these are ones that we might want to be clean for in the future,
# but currently are not
disabled_needs_work = [
    'cppcoreguidelines-use-default-member-init',
    'bugprone-unchecked-optional-access',
    'bugprone-empty-catch',
    'cppcoreguidelines-avoid-const-or-ref-data-members',
    'cppcoreguidelines-pro-type-const-cast',
    'misc-include-cleaner', # warning: useful but quite buggy
    'misc-multiple-inheritance', # public key uses this but should be fixed
    'misc-override-with-different-visibility', # mostly fine but should be looked at
    'modernize-pass-by-value',
    'modernize-use-ranges', # limited by compiler support currently
    'performance-avoid-endl',
    'readability-redundant-typename', # TODO(Botan4) when Clang min version increases this can be fixed
    'readability-convert-member-functions-to-static',
    'readability-inconsistent-declaration-parameter-name', # should fix this, blocked by https://github.com/llvm/llvm-project/issues/60845
    'readability-simplify-boolean-expr', # sometimes ok
    'readability-static-accessed-through-instance',
]

# these we are probably not interested in ever being clang-tidy clean for
disabled_not_interested = [
    '*-array-to-pointer-decay',
    '*-avoid-c-arrays', # triggers also on foo(T x[], size_t len) decls
    '*-else-after-return',
    '*-function-size',
    '*-magic-numbers', # can't stop the magic
    '*-narrowing-conversions',
    '*-no-array-decay',
    '*-use-auto', # not universally a good idea
    '*-use-emplace', # often less clear
    '*-deprecated-headers', # wrong for system headers like stdlib.h
    'cert-dcl21-cpp', # invalid, and removed already in clang-tidy 19
    'bugprone-easily-swappable-parameters',
    'bugprone-implicit-widening-of-multiplication-result', # would be reasonable if it ignored constants (it doesn't)
    'cppcoreguidelines-pro-bounds-avoid-unchecked-container-access',
    'cppcoreguidelines-pro-bounds-pointer-arithmetic',
    'cppcoreguidelines-pro-bounds-constant-array-index',
    'cppcoreguidelines-pro-type-reinterpret-cast',
    'hicpp-signed-bitwise', # would be reasonable if it ignored constants (it doesn't)
    'misc-no-recursion',
    'modernize-avoid-c-style-cast', # the kind of cast clang-tidy is complaining about here is fine
    'modernize-use-trailing-return-type', # fine, but we're not using it everywhere
    'modernize-return-braced-init-list', # thanks I hate it
    'modernize-use-default-member-init',
    'modernize-use-designated-initializers',
    'modernize-use-nodiscard',
    'modernize-use-using', # fine not great
    'readability-avoid-return-with-void-value', # Jack likes doing this
    'readability-function-cognitive-complexity',
    'readability-identifier-length', # lol, lmao
    'readability-math-missing-parentheses',
    'readability-non-const-parameter',
    'readability-use-concise-preprocessor-directives', # it's not more readable...
    'readability-redundant-parentheses', # often improves readability ...
    'readability-redundant-inline-specifier', # Jack likes doing this
    'readability-redundant-access-specifiers', # reneme likes doing this
    'readability-use-anyofallof', # not more readable
]

disabled_checks = sorted(disabled_needs_work + disabled_not_interested)

def create_check_option(enabled, disabled):
    return ','.join(enabled) + ',' + ','.join(['-' + d for d in disabled])

def render_clang_tidy_file(target_dir, enabled, disabled):
    filepath = os.path.join(target_dir, '.clang-tidy')
    print(f'regenerating {filepath}')
    today = datetime.date.today().strftime("%Y-%m-%d")
    with open(filepath, "w", encoding="utf-8") as clang_tidy_file:
        clang_tidy_file.writelines([
            '---\n',
            f'# This file was automatically generated by {sys.argv[0]} --regenerate-inline-config-file on {today}\n',
            '# All manual changes will be lost. Edit the script instead.\n',
            '\n',
            'Checks: >\n'] +
            [ f'    {check},\n' for check in sorted(enabled)] +
            [ f'    -{check},\n' for check in sorted(disabled)] +
            ['---\n'])

def load_compile_commands(build_dir):
    compile_commands_file = os.path.join(build_dir, 'compile_commands.json')
    compile_commands = open(compile_commands_file, encoding='utf8').read()
    return (compile_commands_file, json.loads(compile_commands))

def run_command(cmdline):
    proc = subprocess.Popen(cmdline,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE)

    (stdout, _stderr) = proc.communicate()

    stdout = stdout.decode('utf8')
    # stderr is discarded

    return stdout

def run_clang_tidy(compile_commands_file,
                   check_config,
                   source_file,
                   options):

    cmdline = ['clang-tidy',
               '--quiet',
               '-checks=%s' % (check_config),
               '-p', compile_commands_file,
               '-header-filter=.*']

    if options.fixit:
        cmdline.append('-fix')

    if options.export_fixes_dir:
        os.makedirs(options.export_fixes_dir, exist_ok=True)
        yml_file = os.path.join(options.export_fixes_dir, "%s.yml" % uuid.uuid4())
        cmdline.append("--export-fixes=%s" % yml_file)

    cmdline.append(source_file)

    stdout = run_command(cmdline)

    if stdout != "":
        print("### Errors in", source_file)
        print(stdout)
        sys.stdout.flush()
        return False

    return True

def file_matches(file, args):
    if args is None or len(args) == 0:
        return True

    for arg in args:
        if file.find(arg) > 0:
            return True
    return False

def preproc_file(compile_commands):
    cmd = shlex.split(compile_commands['command'])

    dash_o = cmd.index("-o")
    if dash_o >= 0:
        cmd.pop(dash_o + 1) # remove object file name
        cmd.pop(dash_o)     # remove -o

    if "-c" in cmd:
        cmd.remove("-c")
    cmd.append("-E")

    result = subprocess.run(
        cmd,
        check=True,
        stdout=subprocess.PIPE,
        universal_newlines=True)

    return hashlib.sha256(result.stdout.encode('utf-8')).hexdigest()

def hash_args(**kwargs):
    outer_hash = hashlib.sha256()

    for key in sorted(kwargs.keys()):
        value = kwargs[key]
        outer_hash.update(hashlib.sha256(key.encode('utf-8')).digest())
        outer_hash.update(hashlib.sha256(value.encode('utf-8')).digest())

    return outer_hash.hexdigest()

class CacheDatabase:
    """
    Caches SUCCESSFUL clang-tidy runs on source files to speed up whole-tree
    runs of clang-tidy. The cache key is calculated as the hash of the source file
    and a set of meta information such as 'clang-tidy --version' and the check configuration.
    """
    CACHE_DEBUG = False

    def __init__(self, db_path):
        self.db = sqlite3.connect(db_path)
        cur = self.db.cursor()
        cur.execute("CREATE TABLE IF NOT EXISTS clang_tidy_clean(key UNIQUE)")
        self.db.commit()

    def hit(self, **kwargs):
        cache_key = hash_args(**kwargs)

        cur = self.db.cursor()
        cur.execute("SELECT key from clang_tidy_clean where key = ?", (cache_key, ))
        res = cur.fetchall()

        if len(res) > 0:
            if self.CACHE_DEBUG:
                print("Cache hit for", cache_key)
            return True
        else:
            if self.CACHE_DEBUG:
                print("Cache miss for", cache_key)
            return False

    def record(self, **kwargs):
        cache_key = hash_args(**kwargs)

        cur = self.db.cursor()
        if self.CACHE_DEBUG:
            print("Cache save for", cache_key)
        cur.execute("INSERT OR IGNORE INTO clang_tidy_clean values(?)", (cache_key, ))
        self.db.commit()

    def count(self):
        cur = self.db.cursor()
        cur.execute("SELECT count(*) FROM clang_tidy_clean")
        entries = cur.fetchall()
        return int(entries[0][0])

    def cleanup(self, limit):
        cur = self.db.cursor()

        cur.execute("SELECT key FROM clang_tidy_clean")
        entries = cur.fetchall()

        if len(entries) <= limit:
            return

        to_prune = len(entries) - limit

        # Randomly evict entries if cache has grown too large
        random.shuffle(entries)

        for cache_key in entries[:to_prune]:
            if self.CACHE_DEBUG:
                print("Pruning", cache_key[0])
            cur.execute("DELETE FROM clang_tidy_clean WHERE key = ?", (cache_key[0], ))
        self.db.commit()

def main(args = None): # pylint: disable=too-many-return-statements
    if args is None:
        args = sys.argv

    parser = optparse.OptionParser()

    parser.add_option('-j', '--jobs', action='store', type='int', default=multiprocessing.cpu_count() + 1,
                      help='set number of jobs to run (default %default)')
    parser.add_option('--quiet', action='store_true', default=False)
    parser.add_option('--fixit', action='store_true', default=False)
    parser.add_option('--build-dir', default='build')
    parser.add_option('--list-checks', action='store_true', default=False)
    parser.add_option('--regenerate-inline-config-file', action='store_true', default=False)
    parser.add_option('--only-changed-files', action='store_true', default=False)
    parser.add_option('--only-matching', metavar='REGEX', default='.*')
    parser.add_option('--take-file-list-from-stdin', action='store_true', default=False)
    parser.add_option('--export-fixes-dir', default=None)
    parser.add_option('--cache-db', metavar='DB',
                      help='Path to sqlite3 database file for caching',
                      default=os.getenv("BOTAN_CLANG_TIDY_CACHE"))

    # 100K entries with our current schema is about 15 Mb
    max_db_entries = int(os.getenv("BOTAN_CLANG_TIDY_CACHE_MAX_SIZE") or 100_000)

    (options, args) = parser.parse_args(args)

    if len(args) > 1:
        print("ERROR: Unknown extra arguments to run_clang_tidy.py")
        return 1

    if options.only_changed_files and options.take_file_list_from_stdin:
        print("Cannot use both --only-changed-files and --take-file-list-from-stdin", file=sys.stderr)
        return 1

    files_to_check = []
    if options.only_changed_files:
        changes = run_command(['git', 'diff', '--name-only', '-r', 'master'])

        files_to_check = []
        for file in changes.split():
            if file.endswith('.cpp') or file.endswith('.h'):
                files_to_check.append(os.path.basename(file))

        if len(files_to_check) == 0:
            print("No C++ files were modified vs master, skipping clang-tidy checks")
            return 0
    elif options.take_file_list_from_stdin:
        scan_all = False

        for line in sys.stdin:
            file = os.path.basename(line.strip())
            if file.endswith('.cpp') or file.endswith('.h'):
                files_to_check.append(file)
            elif file in ['run_clang_tidy.py']:
                scan_all = True

        if scan_all:
            # clear this to revert to disable filtering causing all files to be scanned
            files_to_check = []
        elif len(files_to_check) == 0:
            print("No C++ files were provided on stdin, skipping clang-tidy checks")
            return 0

    (compile_commands_file, compile_commands) = load_compile_commands(options.build_dir)

    check_config = create_check_option(enabled_checks, disabled_checks)

    if options.list_checks:
        print(run_command(['clang-tidy', '-list-checks', '-checks', check_config]), end='')
        return 0

    if options.regenerate_inline_config_file:
        render_clang_tidy_file('src', enabled_checks, disabled_checks)
        return 0

    pool = ThreadPool(options.jobs)

    cache = CacheDatabase(options.cache_db) if options.cache_db else None

    start_time = time.time()
    files_checked = 0

    file_matcher = re.compile(options.only_matching)

    clang_tidy_version = None

    if cache is not None:
        if not options.quiet:
            print("Using cache at %s with %d entries" % (options.cache_db, cache.count()))

        results = []

        clang_tidy_version = subprocess.run(['clang-tidy', '--version'],
                                            check=True,
                                            stdout=subprocess.PIPE).stdout.decode('utf-8')

        for info in compile_commands:
            results.append(pool.apply_async(preproc_file, (info, )))

        assert(len(results) == len(compile_commands))
        for (i, result) in enumerate(results):
            compile_commands[i]['preproc'] = result.get()

        if not options.quiet:
            print("Preprocessing/hashing %d files took %.02f sec" % (len(results), time.time() - start_time))
            sys.stdout.flush()

        start_time = time.time()

    results = []
    for info in compile_commands:
        file = info['file']

        if len(files_to_check) > 0 and os.path.basename(file) not in files_to_check:
            continue

        if file_matcher.search(file) is None:
            continue

        cache_hit = False
        if cache is not None:
            if cache.hit(config=check_config, source_file=info['preproc'], clang_tidy=clang_tidy_version):
                cache_hit = True

        if cache_hit and not options.quiet:
            print("Checked", info["file"], " (cached)")
            sys.stdout.flush()

        if not cache_hit:
            files_checked += 1

            results.append((info, pool.apply_async(
                run_clang_tidy,
                (compile_commands_file,
                 check_config,
                 file,
                 options))))

    fail_cnt = 0
    for (info, result) in results:
        success = result.get()

        if not options.quiet:
            print("Checked", info['file'])
            sys.stdout.flush()

        if success and cache is not None:
            cache.record(config=check_config, source_file=info['preproc'], clang_tidy=clang_tidy_version)

        if not success:
            fail_cnt += 1

    time_consumed = time.time() - start_time

    print("Checked %d files in %d seconds" % (files_checked, time_consumed))

    if cache is not None and max_db_entries > 0:
        cache.cleanup(max_db_entries)

    if fail_cnt == 0:
        return 0
    else:
        print("Found clang-tidy errors in %d files" % (fail_cnt))
        return 1

if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python3
# coding=utf8

"""
(C) 2026 Jack Lloyd

Botan is released under the Simplified BSD License (see license.txt)
"""

import argparse
import os
import re
import subprocess
import sys

# Stanza must appear near the start of the file, and must include a date.
GENERATED_RE = re.compile(
    r'This file was automatically generated by\b.*?\bon (\d{4}-\d{2}-\d{2})',
    re.DOTALL)

# Only scan the head of the file; the stanza always lives in the opening
# comment block.
SCAN_BYTES = 512


def run_git(*args):
    return subprocess.check_output(['git'] + list(args), text=True)


def generation_date(content):
    """Return the YYYY-MM-DD stamp from the stanza, or None if no stanza."""
    m = GENERATED_RE.search(content[:SCAN_BYTES])
    return m.group(1) if m else None


def file_at_ref(ref, path):
    """Return the text contents of <ref>:<path>, or None if missing."""
    try:
        return subprocess.check_output(
            ['git', 'show', f'{ref}:{path}'],
            text=True, stderr=subprocess.DEVNULL,
            errors='replace')
    except subprocess.CalledProcessError:
        return None


def file_in_working_tree(path):
    try:
        with open(path, encoding='utf8', errors='replace') as f:
            return f.read(SCAN_BYTES)
    except OSError:
        return None


def changed_files(base):
    """Files changed between base...HEAD plus any uncommitted changes."""
    out = set()
    out.update(run_git('diff', '--name-only', f'{base}...HEAD').split())
    out.update(run_git('diff', '--name-only', 'HEAD').split())
    out.discard('')
    return sorted(out)


def default_base():
    # On GitHub Actions during a pull_request event, the target branch name
    # is in GITHUB_BASE_REF (e.g. "master"). The remote ref is "origin/<name>".
    github_base = os.environ.get('GITHUB_BASE_REF')
    if github_base:
        return f'origin/{github_base}'
    return 'origin/master'


def main(argv=None):
    parser = argparse.ArgumentParser(description="Check that generated files are not modified.")
    parser.add_argument(
        '--base', default=default_base(),
        help='git ref to diff against (default: %(default)s)')
    args = parser.parse_args(argv)

    try:
        run_git('rev-parse', '--verify', args.base)
    except subprocess.CalledProcessError:
        # If the base isn't present (shallow clone, running on master
        # itself, non-PR workflow, ...) there's nothing to compare against.
        print(f'skipping: base ref {args.base!r} not available', file=sys.stderr)
        return 0

    flagged = []
    for path in changed_files(args.base):
        print("Checking '%s'..." % (path))
        working = file_in_working_tree(path)
        if working is None:
            continue  # deleted

        new_date = generation_date(working)
        if new_date is None:
            continue  # not a generated file

        base_content = file_at_ref(args.base, path)
        if base_content is None:
            continue  # newly added file; nothing to compare against

        old_date = generation_date(base_content)
        if old_date is None:
            continue

        if old_date == new_date:
            flagged.append((path, old_date))

    if flagged:
        print('ERROR: generated file(s) were modified without updating the generation date.')
        print('These files are regenerated by a gen_*.py script; edit the script and regenerate')
        print('rather than editing the file directly.')
        print()
        for path, date in flagged:
            print(f'  {path}  (date unchanged: {date})')
        return 1

    return 0


if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python3

"""
(C) 2026 Jack Lloyd

Botan is released under the Simplified BSD License (see license.txt)
"""

import argparse
import configparser
import hashlib
import os
import subprocess
import sys
import tempfile
import urllib.request

def load_config(dep_name):
    config_path = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                               '..', '..', 'configs', 'ci_deps.conf')
    config = configparser.ConfigParser()
    config.read(config_path)

    if dep_name not in config:
        print("Unknown dependency %s - available options are %s" % (
            dep_name, ','.join(config.sections())))
        sys.exit(1)

    section = config[dep_name]
    url = section.get('url')
    sha256 = section.get('sha256')
    if not url or not sha256:
        print("Bad config entry for %s" % (dep_name))
        sys.exit(1)

    return url, sha256

def download(url, fileobj, max_mb):
    max_bytes = max_mb * 1024 * 1024
    req = urllib.request.Request(url)
    hasher = hashlib.sha256()
    total = 0
    with urllib.request.urlopen(req) as resp:
        content_length = resp.headers.get('Content-Length')
        content_length = int(content_length) if content_length is not None else None

        if (content_length is not None) and (content_length > max_bytes):
            raise RuntimeError("Download of %s too large, server reports %d bytes" % (url, content_length))

        while True:
            chunk = resp.read(256 * 1024)
            if not chunk:
                break
            total += len(chunk)
            if (content_length is not None) and (total > content_length):
                raise RuntimeError("Server sent too much data for %s, reported %d" % (url, content_length))
            if total > max_bytes:
                raise RuntimeError("Server sent too much data for %s" % (url))
            hasher.update(chunk)
            fileobj.write(chunk)

    return hasher.hexdigest(), total

def main():
    parser = argparse.ArgumentParser(description='Download a CI dependency with integrity verification')
    parser.add_argument('dep_name', help='Dependency name (section in ci_deps.conf)')
    parser.add_argument('output_path', nargs='?', default=None,
                        help='Output file path (default: filename from URL in current directory)')
    parser.add_argument('--max-download-mb', default=48, type=int,
                        help='Maximum download size in MB')
    parser.add_argument('--extract', default=None, metavar='CMD',
                        help='Extract after download using CMD template (eg "tar -xf {file}")')
    args = parser.parse_args()

    url, expected_sha256 = load_config(args.dep_name)

    if args.extract:
        final_path = None
        fd, tmp_path = tempfile.mkstemp(prefix='ci_dep_')
    else:
        if args.output_path:
            final_path = args.output_path
        else:
            final_path = os.path.basename(urllib.request.url2pathname(url.split('?')[0]))
        fd, tmp_path = tempfile.mkstemp(prefix='.ci_dep_',
                                        dir=os.path.dirname(final_path) or '.')

    try:
        with os.fdopen(fd, 'wb') as f:
            computed_sha256, total = download(url, f, args.max_download_mb)

        if computed_sha256 != expected_sha256:
            print("Checksum failure downloading %s - got %s (%d bytes)" % (
                url, computed_sha256, total))
            return 1

        if args.extract:
            cmd = args.extract.replace('{file}', tmp_path)
            subprocess.run(cmd, shell=True, check=True)
        else:
            os.replace(tmp_path, final_path)
            print(final_path)

        return 0
    except Exception as e:
        print(str(e))
        return 1
    finally:
        try:
            os.unlink(tmp_path)
        except FileNotFoundError:
            pass

if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python3

#
# Strips the KAT harness produced by the Dilithium reference implementation down
# to a less space consuming version. This script was used to generate
# `src/tests/data/pubkey/dilithium_[...].vec` test data from the *.rsp files of
# the reference implementation.
#
# (C) 2022,2023 Jack Lloyd
# (C) 2022 René Meusel, Rohde & Schwarz Cybersecurity
#
# Botan is released under the Simplified BSD License (see license.txt)
#

import sys
import hashlib
import binascii

class KatReader:
    def __init__(self, file):
        self.file = file
        self.last_mlen = None

    def next_value(self):
        while True:
            line = self.file.readline()

            if line == "":
                return (None, None)

            if line.startswith('#') or line == "\n":
                continue

            key, val = line.strip().split(' = ')

            return (key, val)

    def read_kats(self):
        kat = {}

        while True:
            key, val = self.next_value()

            if key is None:
                return # eof

            if key not in ['count', 'seed', 'mlen', 'msg', 'pk', 'sk', 'smlen', 'sm']:
                raise Exception("Unknown key %s" % (key))

            if key in ['count', 'mlen', 'smlen']:
                kat[key] = int(val)
            elif key == 'sm':
                # remove message appended to signature
                kat[key] = val[:-kat['mlen']*2]
            else:
                kat[key] = val

            if key == 'sm':
                yield kat
                kat = {}

def sha3_256(v):
    # v is assumed to be hex
    h = hashlib.sha3_256()
    h.update(binascii.unhexlify(v))
    return h.hexdigest()

def compress_kat(kat):
    first = kat['count'] == 0
    del kat['count']
    del kat['smlen']
    del kat['mlen']

    # rename keys
    kat['Seed'] = kat.pop('seed')
    kat['Msg'] = kat.pop('msg')

    kat['HashPk'] = sha3_256(kat.pop('pk'))
    kat['HashSk'] = sha3_256(kat.pop('sk'))

    sig = kat.pop('sm')
    if first:
        kat['Sig'] = sig
    kat['HashSig'] = sha3_256(sig)

    return kat

def map_mode(mode):
    if mode == 'Dilithium2':
        return '4x4'
    if mode == 'Dilithium2-AES':
        return '4x4_AES'
    if mode == 'Dilithium3':
        return '6x5'
    if mode == 'Dilithium3-AES':
        return '6x5_AES'
    if mode == 'Dilithium5':
        return '8x7'
    if mode == 'Dilithium5-AES':
        return '8x7_AES'

    raise Exception('Unknown Dilithium mode', mode)

def main(args = None):
    if args is None:
        args = sys.argv

    randomized = True
    is_mldsa = True

    type = 'Randomized' if randomized else 'Deterministic'
    name = 'ml-dsa' if is_mldsa else 'dilithium'

    for file in args[1:]:
        mode = map_mode(open(file).readline().strip()[2:])

        reader = KatReader(open(file))

        output = open('src/tests/data/pubkey/%s_%s_%s.vec' % (name, mode, type), 'w')

        print("# See src/scripts/dev_tools/gen_dilithium_kat.py\n", file=output)
        print("[%s_%s]" % (name.upper() if is_mldsa else name.capitalize(), mode), file=output)

        for kat in reader.read_kats():
            kat = compress_kat(kat)

            for key in ['Seed', 'Msg', 'HashPk', 'HashSk', 'Sig', 'HashSig']:
                if key in kat:
                    print(key, '=', kat[key], file=output)
            print("\n", file=output)

        output.close()

if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python3

# Script to run inside the CI container to test the botan
# TLS client/server with TLS-Anvil
#
# (C) 2023,2026 Jack Lloyd
# (C) 2023 Fabian Albert, Rohde & Schwarz Cybersecurity
#
# Botan is released under the Simplified BSD License (see license.txt)
import sys
import argparse
import os
import subprocess

sys.path.insert(0, os.path.join(os.path.dirname(os.path.abspath(__file__)), '..'))
from repo_config import RepoConfig  # noqa: E402

_repo_config = RepoConfig()


class Config:
    """ Hardcoded configurations for this CI script """
    tls_anvil_docker_image = _repo_config["TLS_ANVIL_DOCKER_IMAGE"]
    tls_anvil_version_tag = _repo_config["TLS_ANVIL_VERSION_TAG"]
    key_and_cert_storage_path = "/tmp/"
    test_suite_results_dest = "."
    test_suite_results_dir_name = "TestSuiteResults"
    tmp_key_file_name = "tmp_rsa_key.pem"
    tmp_cert_file_name = "tmp_rsa_cert.pem"
    server_dest_ip = "127.0.0.1"
    server_dest_port = 4433
    client_test_port = 4433
    trigger_server_port = 8090
    botan_server_log = "./logs/botan_server.log"
    botan_client_log_dir = "./logs/botan_client"
    anvil_policy_file = "src/scripts/tls_anvil/anvil_policy.txt"

def group_output(group_title: str, func):
    """
    Wraps a function to be called within a GitHub actions group, so that
    the console output is expandable.

    Returns the wrapped function
    """
    def wrapped_func(*args, **kwargs):
        print(f"::group::{group_title}", flush=True)
        ret = func(*args, **kwargs)
        print("\n::endgroup::", flush=True)
        return ret
    return wrapped_func


def create_cert_and_key(botan_cli):
    """
    Create a X.509 certificate and associated RSA key at Config.key_and_cert_storage_path
    using Botan's CLI.

    Returns: (<cert path>, <key path>)
    """

    key_path = os.path.join(Config.key_and_cert_storage_path, Config.tmp_key_file_name)
    cert_path = os.path.join(Config.key_and_cert_storage_path, Config.tmp_cert_file_name)

    with open(key_path, 'w', encoding='utf-8') as keyfile:
        subprocess.run([botan_cli, "keygen", "--algo=RSA", "--params=2048"], stdout=keyfile, check=True)

    with open(cert_path, 'w', encoding='utf-8') as certfile:
        subprocess.run([botan_cli, "gen_self_signed", key_path, "localhost"], stdout=certfile, check=True)

    return (cert_path, key_path)


def server_test(botan_cli: str, parallel: int):
    """ Test the Botan TLS server """
    cert_path, key_path = create_cert_and_key(botan_cli)
    docker_img = f"{Config.tls_anvil_docker_image}{Config.tls_anvil_version_tag}"

    group_output("Pull TLS-Anvil image", subprocess.run)(["docker", "pull", docker_img], check=True)

    tls_anvil_cmd = [
        "docker", "run",
        "--network", "host",
        "-v", f"{Config.test_suite_results_dest}:/output",
        docker_img,
        "-strength", "1",
        "-parallelHandshakes", str(parallel),
        "-disableTcpDump",
        "-outputFolder", os.path.join(Config.test_suite_results_dest, Config.test_suite_results_dir_name),
        "-connectionTimeout", "5000",
        "server", "-connect", f"{Config.server_dest_ip}:{Config.server_dest_port}"
    ]

    botan_server_cmd = [
        botan_cli, "tls_http_server", cert_path, key_path,
        f"--port={Config.server_dest_port}",
        f"--policy={Config.anvil_policy_file}",
    ]

    os.makedirs(os.path.dirname(Config.botan_server_log), exist_ok=True)

    # Run Botan and test is with TLS-Anvil
    with open(Config.botan_server_log, 'w', encoding='utf-8') as server_log_file:
        botan_server_process = subprocess.Popen(botan_server_cmd, stdout=server_log_file, stderr=server_log_file)
        subprocess.run(tls_anvil_cmd, check=True)
        botan_server_process.kill()


def client_test(botan_cli: str, parallel: int):
    """ Test the Botan TLS client """
    docker_img = f"{Config.tls_anvil_docker_image}{Config.tls_anvil_version_tag}"

    group_output("Pull TLS-Anvil image", subprocess.run)(["docker", "pull", docker_img], check=True)

    trigger_server_script = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "tls_anvil_trigger_server.py")

    trigger_server_cmd = [
        "python3", trigger_server_script,
        "--botan-cli", botan_cli,
        "--tls-anvil-policy", Config.anvil_policy_file,
        "--tls-anvil-host", Config.server_dest_ip,
        "--tls-anvil-port", str(Config.client_test_port),
        "--listen-port", str(Config.trigger_server_port),
        "--log-dir", Config.botan_client_log_dir,
    ]

    tls_anvil_cmd = [
        "docker", "run",
        "--network", "host",
        "-v", f"{Config.test_suite_results_dest}:/output",
        docker_img,
        "-strength", "1",
        "-parallelHandshakes", str(parallel),
        "-disableTcpDump",
        "-outputFolder", os.path.join(Config.test_suite_results_dest, Config.test_suite_results_dir_name),
        "-connectionTimeout", "5000",
        "client",
        "-port", str(Config.client_test_port),
        "-triggerScript", "curl", "--connect-timeout", "2",
        f"http://{Config.server_dest_ip}:{Config.trigger_server_port}/trigger"
    ]

    os.makedirs(Config.botan_client_log_dir, exist_ok=True)

    trigger_server_process = subprocess.Popen(trigger_server_cmd)
    try:
        subprocess.run(tls_anvil_cmd, check=True)
    finally:
        trigger_server_process.terminate()
        trigger_server_process.wait()


def main(args=None):
    if args is None:
        args = sys.argv[1:]

    parser = argparse.ArgumentParser()
    parser.add_argument("--botan-cli", help="Path to botan CLI executable", required=True)
    parser.add_argument("--test-target", help="The TLS side to test", choices=['client', 'server'], required=True)
    parser.add_argument("--parallel", help="The number of parallel handshakes", type=int, default=16)

    args = vars(parser.parse_args(args))

    botan_cli = args["botan_cli"]
    if not os.access(botan_cli, os.X_OK):
        raise FileNotFoundError(f"Botan CLI not found or not executable: '{botan_cli}'")

    if args["test_target"] == "server":
        server_test(botan_cli, args["parallel"])
    elif args["test_target"] == "client":
        client_test(botan_cli, args["parallel"])

    return 0


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python3

"""
(C) 2016,2025 Jack Lloyd
(C) 2017 Fabian Weissberg, Rohde & Schwarz Cybersecurity

Botan is released under the Simplified BSD License (see license.txt)

NOTE: This script requires the Jinja templating library to be installed.
"""

import sys
import datetime
import re
from jinja2 import Environment, FileSystemLoader

# This must match OID::hash_code
def hash_oid(oid):
    word_size = 2 ** 64
    h = 0x621F302327D9A49A

    for part in map(int, oid.split('.')):
        h = (h * 193) % word_size
        h += part

    # This reduction step occurs in static_oids.cpp.in
    return h % 858701

# This must match hash_oid_name in static_oids.cpp.in
def hash_oid_name(name):
    word_size = 2 ** 64

    h = 0x8188B31879A4879A

    for part in map(ord, name):
        h = (h * 251) % word_size
        h += part

    return h % 805289

def format_oid(oid):
    return '{' + oid.replace('.', ', ') + '}'

def render_static_oid(m):
    res = []

    name_hashes = {}
    oid_hashes = {}

    for (k, v) in m.items():

        # Verify no collisions between any of the values
        oid_hc = hash_oid(v)
        if oid_hc in oid_hashes:
            raise Exception("Hash collision between %s and %s" % (v, oid_hashes[oid_hc]))
        oid_hashes[oid_hc] = v

        name_hc = hash_oid_name(k)
        if name_hc in name_hashes:
            raise Exception("Hash collision between %s and %s" % (k, name_hashes[name_hc]))
        name_hashes[name_hc] = k

        res.append({ 'oid_hash': oid_hc,
                     'name_hash': name_hc,
                     'name': k,
                     'oid': format_oid(v) })

    return res

def format_oid_with_name(m):
    return [ {'name': kv[0], 'oid': format_oid(kv[1])} for kv in m ]

def main(args = None):
    """
    Regenerate src/lib/asn1/static_oids.cpp
    """
    if args is None:
        args = sys.argv

    oid_lines = open('./src/build-data/oids.txt', encoding='utf8').readlines()

    oid_re = re.compile(r"^([0-9][0-9.]+) += +([A-Za-z0-9_\./\(\), -]+)$")
    hdr_re = re.compile(r"^\[([a-z0-9_]+)\]$")

    oid2str = {}
    str2oid = {}
    dup_oids = []
    aliases = []

    for line in oid_lines:
        line = line.strip()
        if len(line) == 0:
            continue

        if line[0] == '#':
            continue

        match = hdr_re.match(line)
        if match is not None:
            continue

        match = oid_re.match(line)
        if match is None:
            raise Exception(line)

        oid = match.group(1)
        name = match.group(2)

        if name not in str2oid and oid not in oid2str:
            str2oid[name] = oid
            oid2str[oid] = name
        elif name in str2oid:
            dup_oids.append((name, oid))
        elif oid in oid2str:
            aliases.append((name, oid))

    this_script = sys.argv[0]
    date = datetime.date.today().strftime("%Y-%m-%d")

    env = Environment(loader=FileSystemLoader("src/build-data/templates"))

    with open('./src/lib/asn1/static_oids.cpp', encoding='utf8', mode='w') as static_oids:
        template = env.get_template("static_oids.cpp.in")
        static_oids.write(template.render(script=this_script,
                                          date=date,
                                          static_oid_data=render_static_oid(str2oid),
                                          dup_oids=format_oid_with_name(dup_oids),
                                          aliases=format_oid_with_name(aliases)))
        static_oids.write("\n")

    return 0

if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python3

"""
(C) 2023 Jack Lloyd

Botan is released under the Simplified BSD License (see license.txt)
"""

import sys
import json
import re

def map_group_name(group):
    if group.startswith('brainpoolP'):
        return group.replace('P', '')

    return group

def main(args = None):
    if args is None:
        args = sys.argv

    brainpool_t = re.compile('brainpoolP[0-9]+t1')

    last_group = None
    last_msg = None

    for file in args[1:]:
        if file.find('p1363') > 0:
            continue

        if file.find('webcrypto') > 0:
            continue

        data = json.loads(open(file).read())
        assert(data['algorithm'] == 'ECDSA')

        for group in data['testGroups']:
            if brainpool_t.match(group['key']['curve']):
                continue

            if last_group != group['key']['curve']:
                print("Group =", map_group_name(group['key']['curve']))
                last_group = group['key']['curve']
            print("Hash =", group['sha'])
            print()

            print("Px = 0x%s" % (group['key']['wx']))
            print("Py = 0x%s" % (group['key']['wy']))

            for test in group['tests']:
                print("\n# Test %d (%s)" % (test['tcId'], test['comment']))

                if last_msg != test['msg']:
                    print("Msg =", test['msg'])
                    last_msg = test['msg']
                print("Signature =", test['sig'])

                accept = test['result'][0].upper() == 'V'

                # We force this to V since all other "Acceptable" signatures
                # in the test set are ones we do want to reject, eg due to
                # invalid DER encodings
                if(test['comment'] == 'Hash weaker than DL-group'):
                    accept = True
                print("Valid =", 1 if accept else 0)

            print()

if __name__ == '__main__':
    sys.exit(main())



# (C) 2025 Jack Lloyd
#     2025 René Meusel, Rohde & Schwarz Cybersecurity
#
# Botan is released under the Simplified BSD License (see license.txt)
#

"""
User-defined GDB CLI command to help in testing GCC's stack scrubbing
annotations. See also src/scripts/test_strubbed_symbols.py for details.

This can be used in interactive GDB sessions for debugging and development.

   $> gdb -x src/scripts/gdb/strubtest.py ./botan
   (gdb) strubtest Botan::SHA_256::compress_digest
   strubtest setup done
   (gdb) run hash --algo=SHA-256 readme.rst
   Success: stackframe of Botan::SHA_256::compress_digest contains 10176 zero bytes after invocation
   ECD8814EDC180601831FDAD7DEDCF24D57F9F002295346E7D558C941AD318F8E readme.rst
"""

import gdb

def stack_pointer(frame):
    return frame.read_register('sp')

def frame_pointer(frame):
    return frame.read_register('fp')

def current_stackframe_memory_span(frame):
    start, end = stack_pointer(frame), frame_pointer(frame)
    if frame.architecture().name() == "aarch64" and end < start:
        start, end = end, start # On aarch64, the stack grows downwards!
    return (start, end - start)

def report_error(error):
    gdb.write(f"Error: {error}\n", gdb.STDERR)

def report_status(status):
    gdb.write(f"{status}\n", gdb.STDOUT)

class PostStrubLocation(gdb.FinishBreakpoint):
    """
    This (temporary) breakpoint is placed by the StrubTarget at the end of the
    "virtual wrapper" GCC introduced to scrub the target's stackframe. When hit
    it validates that the this stackframe indeed contains zero bytes only.
    """

    def __init__(self, caller_frame, stackframe, function_name):
        super().__init__(caller_frame, internal=True)
        self.function_name = function_name
        self.payload_stack_memory = stackframe

    def stackframe_memory(self):
        return gdb.selected_inferior().read_memory(*self.payload_stack_memory)

    def stackframe_size(self):
        return self.payload_stack_memory[1]

    def is_stackframe_scrubbed(self):
        return all(b'\x00' == b for b in self.stackframe_memory())

    def stop(self):
        if self.stackframe_size() == 0:
            report_error(f"{self.function_name} has an empty stackframe, cannot validate")
        elif not self.is_stackframe_scrubbed():
            report_error(f"{self.function_name} didn't get its stack frame scrubbed after usage")
        else:
            report_status(f"Success: stackframe of {self.function_name} contains {self.stackframe_size()} zero bytes after invcoation")

class TargetReturnLocation(gdb.Breakpoint):
    """
    This special breakpoint finds one or more return instructions in the target
    frame and registers itself at those instruction addresses. When hit, it will
    obtain the stackframe size just before the function returns and set a
    PostStrubLocation breakpoint at the caller frame.
    """

    @staticmethod
    def find_and_register_in(frame, function_name):
        arch = frame.architecture()
        assert arch.name() == "aarch64", "TargetReturnLocation is meant for aarch64"
        disass = arch.disassemble(frame.block().start, frame.block().end)
        addrs = [f"0x{instr['addr']:x}" for instr in disass if instr['asm'] == 'ret']
        if not addrs:
            report_error(f"no ret instructions found in {function_name}")
        else:
            for ret_address in addrs:
                TargetReturnLocation(ret_address, function_name)

    def __init__(self, address, function_name):
        super().__init__(f"*{address}", internal=True, temporary=True)
        self.function_name = function_name

        # Workaround: gdb.Breakpoint has a temporary= param in its constructor,
        # that is meant to delete the breakpoint after it has been hit. Though,
        # it didn't work for some reason.
        self.hit = False

    def stop(self):
        if not self.hit:
            target_frame = gdb.newest_frame()
            caller_frame = target_frame.older()
            stackframe = current_stackframe_memory_span(target_frame)
            PostStrubLocation(caller_frame, stackframe, self.function_name)
            self.hit = True

class StrubTarget(gdb.Breakpoint):
    """
    This special breakpoint shall be set to a symbol that was marked with
    GCC's __attribute__(strub("internal")). When hit, it will note the size of
    its stackframe and set another temporary breakpoint at the end of the
    "virtual wrapper" GCC introduced, see PostStrubLocation. There, we'll check
    if the now-invalidated stackframe indeed contains only zero bytes.
    """

    def __init__(self, function_name):
        super().__init__(function_name, internal=True)
        self.function_name = function_name

    def stop(self):
        target_frame = gdb.newest_frame()
        caller_frame = gdb.newest_frame().older()

        # __attribute__( strub("internal") ) creates a "virtual wrapper" around
        # the annotated function. This wrapper has the same symbol name as the
        # actual target function. To tell them apart, we simply _assume_ that
        # the wrapper is always at the lower address in the binary. This may
        # change in the future or differ across compiler versions!
        if len(self.locations) > 1:
            wrapper_address = min(loc.address for loc in self.locations)
            if wrapper_address == target_frame.pc():
                return False

        if target_frame.architecture().name() == "aarch64":
            # On aarch64, the stackframe size is not available at the beginning
            # of the function, so we set breakpoints at the return instructions
            # of the current frame and obtain the stackframe size there.
            TargetReturnLocation.find_and_register_in(target_frame, self.function_name)
        else:
            # On other platforms (e.g. x86_64), we can directly obtain the
            # stackframe size at the beginning of the function and set the
            # PostStrubLocation breakpoint immediately.
            stackframe = current_stackframe_memory_span(target_frame)
            PostStrubLocation(caller_frame, stackframe, self.function_name)

        # Don't stop for interactive inspection at this location
        return False

class StrubTest(gdb.Command):
    """
    User-defined gdb command to set up a stack scrubbing (strub) test allowing
    to check that the invocation of a given symbol gets its stack cleared after
    returning. Stack scrubbing is a feature in GCC 14 and newer and is enabled
    for relevant functions using `./configure.py --enable-stack-scrubbing`.
    """

    def __init__(self):
        super().__init__("strubtest", gdb.COMMAND_USER)

    def invoke(self, argstring, _):
        if not argstring:
            report_error("a target symbol name is required")
            return

        bp = StrubTarget(argstring)
        if bp.pending:
            bp.delete()
            report_error(f"the provided symbol '{argstring}' does not appear to be available")

        report_status("strubtest setup done")

# Register the custom command with GDB. It may be invoked from GDB's CLI:
#
#  strubtest <annotated symbol, e.g. "Botan::SHA_256::compress_digest">
#
# Any time a test program (compiled with --enable-stack-scrubbing --debug-mode)
# invokes this symbol GDB will either print "Success: ..." or "Error: ...",
# depending on the outcome of the test.
StrubTest()



#!/usr/bin/env python3

"""
Automatically generate declarations for the FFI layer

(C) 2019, 2023 Jack Lloyd

Botan is released under the Simplified BSD License (see license.txt)
"""

import traceback
from pycparser import c_ast, parse_file

ffi_header = 'src/lib/ffi/ffi.h'

def to_ctype(typ, is_ptr):

    if typ.startswith('botan_') and typ.endswith('_t'):
        return 'c_void_p'

    if typ == 'botan_view_ctx':
        return 'c_void_p'

    if typ == 'botan_view_bin_fn':
        return '_VIEW_BIN_CALLBACK'
    if typ == 'botan_view_str_fn':
        return '_VIEW_STR_CALLBACK'

    if is_ptr is False:
        if typ == 'uint32':
            return 'c_uint32'
        elif typ == 'size_t':
            return 'c_size_t'
        elif typ == 'uint8_t':
            return 'c_uint8'
        elif typ == 'uint32_t':
            return 'c_uint32'
        elif typ == 'uint64_t':
            return 'c_uint64'
        elif typ == 'int':
            return 'c_int'
        elif typ == 'unsigned':
            return 'c_uint'
    else:
        if typ == 'void':
            return 'c_void_p'
        elif typ in ['char', 'uint8_t']: # hack
            return 'c_char_p'
        elif typ == 'size_t':
            return 'POINTER(c_size_t)'
        #elif typ == 'uint8_t':
        #    return 'POINTER(c_uint8)'
        elif typ == 'uint32_t':
            return 'POINTER(c_uint32)'
        elif typ == 'uint64_t':
            return 'POINTER(c_uint64)'
        elif typ == 'int':
            return 'POINTER(c_int)'

    raise Exception("Unknown type %s/%d" % (typ, is_ptr))

class FuncDefVisitor(c_ast.NodeVisitor):
    def __init__(self):
        self.fn_decls = []

    def emit_decls(self):
        decls = ''
        for (fn_name, fn_args) in self.fn_decls:
            decl = "    ffi_api(dll.%s," % (fn_name)
            if len(fn_args) > 4:
                decl += "\n            "
            else:
                decl += ' '

            decl += '[' + ', '.join(fn_args) + '])\n'

            decls += decl
        return decls

    def visit_FuncDecl(self, node):

        if not isinstance(node.type, c_ast.TypeDecl):
            #print("ignoring", node.type)
            return

        if node.type.type.names != ['int']:
            #print("ignoring", node.type)
            return

        # all functions returning ints:
        fn_name = node.type.declname

        ignored = [
            'botan_view_bin_fn',
            'botan_view_str_fn',
            'botan_same_mem', # deprecated
            'botan_mceies_encrypt', # dead
            'botan_mceies_decrypt', # dead
            'botan_rng_init_custom', # fixme
        ]
        if fn_name in ignored:
            return

        fn_args = []

        try:
            for param in node.args.params:

                is_ptr = False
                typ = None

                if isinstance(param.type, c_ast.PtrDecl):
                    is_ptr = True

                    if isinstance(param.type.type, c_ast.PtrDecl):
                        typ = param.type.type.type.type.names[0]
                    else:
                        typ = param.type.type.type.names[0]
                elif isinstance(param.type, c_ast.ArrayDecl):
                    is_ptr = True
                    typ = param.type.type.type.names[0]
                else:
                    typ = param.type.type.names[0]

                ctype = to_ctype(typ, is_ptr)
                fn_args.append(ctype)

            self.fn_decls.append((fn_name, fn_args))
        except Exception as e:
            print(traceback.format_exc())
            print("FAILED for '%s': %s" % (fn_name, e))

ast = parse_file(ffi_header, use_cpp=True, cpp_args=['-Ibuild/include', '-std=c89', '-DBOTAN_DLL=', '-DBOTAN_NO_DEPRECATED_WARNINGS'])
#print(ast)
v = FuncDefVisitor()
v.visit(ast)

#print("\n".join(v.fn_decls))
print(v.emit_decls())



#!/usr/bin/env python3

"""
(C) 2025 Jack Lloyd
Botan is released under the Simplified BSD License (see license.txt)

Compare two JSON files output by `botan speed --format=json` and report
on noticeable improvements or regressions in performance.
"""

import json
import optparse # pylint: disable=deprecated-module
import sys
import re

def ops_per_second(events, nanos):
    return (events * 1000000000) / nanos

def format_pct(r):
    #assert r > 1
    return "%.01f%%" % (abs(r - 1) * 100)

def parse_perf_report(report):
    if len(report) == 0:
        print("No report data")
        return None

    version = {'version': 'unknown', 'git': 'unknown'}
    if 'version' in report[0]:
        version = report[0]

        if 'git' in version and version['git'] != 'unknown':
            version['git'] = version['git'][:12]

        report = report[1:]

    re_with_suffix = re.compile(r'(.*) \[[a-z0-9_]+\]$')

    results = []
    for t in report:
        if 'algo' in t and 'op' in t and 'events' in t and 'nanos' in t:

            algo = t['algo']
            match = re_with_suffix.match(algo)
            if match:
                algo = match.group(1)
            op = t['op']
            if 'buf_size' in t:
                op += ' ' + str(t['buf_size']) + ' buffer'

            results.append(((algo, op), ops_per_second(t['events'], t['nanos'])))
        else:
            print("Unexpected record", t)

    results = sorted(results, key=lambda r: r[0])
    return (version, results)

def read_src(src):
    if src in ['stdin', '-']:
        return sys.stdin.read()
    else:
        return open(src, encoding='utf8').read()

def main(args = None):
    if args is None:
        args = sys.argv

    usage = "usage: %prog [options] base.json compare.json"
    parser = optparse.OptionParser(usage=usage)

    parser.add_option('--limit', default=3, help="set reporting limit (as percent)")
    parser.add_option('--filter', default='.*', help="filter results by regex")

    (options, args) = parser.parse_args(args)

    if len(args) != 3:
        print("Usage: compare_perf.py orig.json new.json")
        return 1

    src1 = args[1]
    src2 = args[2]

    if src1 == src2:
        print("Doesn't make sense to compare the same inputs")
        return 1

    (ver0, rep0) = parse_perf_report(json.loads(read_src(src1)))
    (ver1, rep1) = parse_perf_report(json.loads(read_src(src2)))

    reportable = float(options.limit) / 100
    filter_re = re.compile(options.filter)

    def diff(k, a, b):
        return k in a and k in b and a[k] != b[k] and (a[k] != "unknown" or b[k] != "unknown")

    if ver0 and ver1:
        s = "Diff between "

        diff_version = diff('version', ver0, ver1)
        diff_git = diff('git', ver0, ver1)

        # TODO check/diff the compiler flags

        s += src1

        if diff_version or diff_git:
            s += " ("
            if diff_version and diff_git:
                s += ver0['version'] + " " + ver0['git']
            elif diff_version:
                s += ver0['version']
            elif diff_git:
                s += ver0['git']
            s += ")"

        s += " and " + src2

        if diff_version or diff_git:
            s += " ("
            if diff_version and diff_git:
                s += ver1['version'] + " " + ver1['git']
            elif diff_version:
                s += ver1['version']
            elif diff_git:
                s += ver1['git']
            s += ")"

        s += "\n"

        print(s)

    data_points = 0
    speedups = []
    slowdowns = []
    missing = 0

    while rep0 != [] and rep1 != []:
        while rep0 != [] and rep1 != [] and rep0[0][0] != rep1[0][0]:
            missing += 1
            if rep0[0][0] < rep1[0][0]:
                rep0 = rep0[1:]
            else:
                rep1 = rep1[1:]

        if rep0 != [] and rep1 != []:
            assert rep0[0][0] == rep1[0][0]
            algo = ' '.join(rep0[0][0])

            if filter_re.search(algo) is not None:
                orig = rep0[0][1]
                new = rep1[0][1]

                ratio = new / orig

                data_points += 1

                if ratio >= 1 + reportable:
                    speedups.append((ratio, algo))
                elif (orig / new) >= 1 + reportable:
                    slowdowns.append((orig / new, algo))

            # go to next
            rep0 = rep0[1:]
            rep1 = rep1[1:]

    for (pct, algo) in sorted(speedups, key=lambda v: v[0], reverse=True):
        print("+ %s improvement in %s" % (format_pct(pct), algo))

    for (pct, algo) in sorted(slowdowns, key=lambda v: v[0]):
        print("- %s regression in %s" % (format_pct(pct), algo))

    if data_points > 0:
        print("\nSummary: over %d tests saw %d speedups and %d slowdowns" % (data_points, len(speedups), len(slowdowns)))
    else:
        print("\nNo data points")

    if missing > 0:
        print("NOTE: there were %d data points in one set but not the other" % (missing))

if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python3

import sys
import datetime

# (C) 2011,2014,2015,2016 Jack Lloyd
# Botan is released under the Simplified BSD License (see license.txt)

# Used to generate src/lib/math/mp/mp_comba.cpp

def comba_indexes(N):

    indexes = []

    for i in range(0, 2*N):
        x = []

        for j in range(max(0, i-N+1), min(N, i+1)):
            x += [(j,i-j)]
        indexes += [sorted(x)]

    return indexes

def comba_sqr_indexes(N):

    indexes = []

    for i in range(0, 2*N):
        x = []

        for j in range(max(0, i-N+1), min(N, i+1)):
            if j < i-j:
                x += [(j,i-j)]
            else:
                x += [(i-j,j)]
        indexes += [sorted(x)]

    return indexes

def comba_multiply_code(N):
    indexes = comba_indexes(N)

    for (i,idx) in zip(range(0, len(indexes)), indexes):
        for pair in idx:
            print("   accum.mul(x[%d], y[%d]);" % (pair[0], pair[1]))

        print("   z[%d] = accum.extract();" % (i))

def comba_square_code(N):
    indexes = comba_sqr_indexes(N)

    for (rnd,idx) in zip(range(0, len(indexes)), indexes):
        for (i,pair) in zip(range(0, len(idx)), idx):
            if pair[0] == pair[1]:
                print("   accum.mul(x[%d], x[%d]);" % (pair[0], pair[1]))
            elif i % 2 == 0:
                print("   accum.mul_x2(x[%d], x[%d]);" % (pair[0], pair[1]))

        print("   z[%d] = accum.extract();" % (rnd))

def main(args = None):
    if args is None:
        args = sys.argv

    if len(args) <= 1:
        sizes = [4, 6, 7, 8, 9, 16, 24]
    else:
        sizes = map(int, args[1:])

    print("""/*
* Comba Multiplication and Squaring
*
* This file was automatically generated by %s on %s
* All manual changes will be lost. Edit the script instead.
*
* Botan is released under the Simplified BSD License (see license.txt)
*/

#include <botan/internal/mp_core.h>

namespace Botan {
""" % (sys.argv[0], datetime.date.today().strftime("%Y-%m-%d")))

    for n in sizes:
        print("/*\n* Comba %dx%d Squaring\n*/" % (n, n))
        print("void bigint_comba_sqr%d(word z[%d], const word x[%d]) {" % (n, 2*n, n))
        print("   word3<word> accum;\n")

        comba_square_code(n)

        print("}\n")

        print("/*\n* Comba %dx%d Multiplication\n*/" % (n, n))
        print("void bigint_comba_mul%d(word z[%d], const word x[%d], const word y[%d]) {" % (n, 2*n, n, n))
        print("   word3<word> accum;\n")

        comba_multiply_code(n)

        print("}\n")

    print("}  // namespace Botan")

    return 0

if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python3

# Parses a TLS-Anvil results directory. Returns 0 iff all results are expected.
#
# (C) 2023,2026 Jack Lloyd
# (C) 2023 Fabian Albert, Rohde & Schwarz Cybersecurity
#
# Botan is released under the Simplified BSD License (see license.txt)
import sys
import argparse
import os
import json
import logging


result_level = {
    "STRICTLY_SUCCEEDED": 0,
    "CONCEPTUALLY_SUCCEEDED": 1,
    "PARTIALLY_FAILED": 2,
    "FULLY_FAILED": 3,
}


def xfail_list(side):
    """Return list of tests that are expected to fail"""

    conceptually_succeeded = {
        # Okay: RFC does not specifically define an alert. Bogo Test expects an DecodeError Alert
        #   while TLS-Anvil expects an IllegalParameter Alert. We use the DecodeError Alert.
        "server.tls13.rfc8446.PreSharedKey.isLastButDuplicatedExtension",
    }

    partially_failed = {
        # We accept the TLS 1.2 brainpool curve IDs in 1.3
        "server.tls13.rfc8446.KeyShare.serverAcceptsDeprecatedGroups",
    }

    fully_failed = {
        # If ClientHello has no extensions that includes supported signatures; TLS 1.2 then
        # requires us to treat that as equivalent to supporting only SHA-1 -- which we do not support
        "server.tls12.rfc5246.ClientHello.leaveOutExtensions",

        # TLS-Anvil expects us to tolerate any legacy_version even ones that RFC 8446 explicitly
         # states we MUST reject - https://github.com/tls-attacker/TLS-Anvil/issues/60
        # 8446-bis has a change specifically mandating what we do right now
        "client.tls13.rfc8446.SupportedVersions.invalidLegacyVersion",

        # TLS-Anvil seems to assume KeyUpdate response is immediate rather than opportunistic
        # Possibly this can be fixed by using a dedicated util rather than tls_client
        "both.tls13.rfc8446.KeyUpdate.respondsWithValidKeyUpdate",
        "both.tls13.rfc8446.KeyUpdate.appDataUnderNewKeysSucceeds",

        # We accept the TLS 1.2 brainpool curve IDs in 1.3
        "server.tls13.rfc8446.KeyShare.serverAcceptsDeprecatedGroupsAllAtOnce",
    }

    if side == 'client':
        # TLS-Anvil's scanning phase seems to have a bug and decides we don't support handshake fragmentation
        # Works for server side though
        fully_failed.add("both.tls12.rfc5246.Fragmentation.recordFragmentationSupported")

    xfails = {}
    for test in conceptually_succeeded:
        if test.startswith('both.') or test.startswith(side):
            xfails[test] = result_level["CONCEPTUALLY_SUCCEEDED"]
    for test in partially_failed:
        if test.startswith('both.') or test.startswith(side):
            xfails[test] = result_level["PARTIALLY_FAILED"]
    for test in fully_failed:
        if test.startswith('both.') or test.startswith(side):
            xfails[test] = result_level["FULLY_FAILED"]
    return xfails


def extract_method_id(json_data):
    """Extract the method_id from a test result JSON."""
    return (json_data["TestClass"] + "." + json_data["TestMethod"]).removeprefix("de.rub.nds.tlstest.suite.tests.")


def failing_test_info(json_data, method_id, expected_label) -> str:
    """ Print debug information about a failing test """
    info_str = ""
    try:
        method_class, method_name = method_id.rsplit('.', 1)
        info = [f"Error: {method_id} - Unexpected result '{json_data['Result']}' (expected {expected_label})"]
        info += [""]
        info += [f"Class Name: 'de.rub.nds.tlstest.suite.tests.{method_class}'"]
        info += [f"Method Name: '{method_name}'"]
        info += [""]

        metadata = json_data.get("MetaData") or {}
        rfc = metadata.get("rfc")
        if rfc is not None:
            info += [f"RFC {rfc.get('number', '?')}, Section {rfc.get('section', '?')}:"]
        else:
            info += ["Custom Test Case:"]

        description = metadata.get("description", "")
        if description:
            info += [description]
        info += [""]

        info += [f"Result: {json_data['Result']}"]

        if json_data.get('DisabledReason'):
            info += [f"Disabled Reason: {json_data['DisabledReason']}"]
        if json_data.get('FailedReason'):
            info += [f"Failed Reason: {json_data['FailedReason']}"]

        info += [""]
        info_str = "\n".join(info)

        # Color in red
        info_str = "\n".join([f"\033[0;31m{line}\033[0m" for line in info_str.split("\n")])

        # In GitHub Actions logging group
        info_str = f"::group::{info_str}\n::endgroup::"

    except (KeyError, TypeError):
        logging.warning("Cannot process test info for %s", method_id)
        info_str = f"Error: {method_id} - Unexpected result '{json_data.get('Result', '?')}'"

    return info_str


def process_test_result(result_path: str, xfails: dict, seen_xfails: set):
    """
    Given a path, process the respective test result .json file.
    Returns True iff the results are expected.
    """
    success = False
    with open(result_path, "r", encoding="utf-8") as f:
        try:
            json_data = json.load(f)
            method_id = extract_method_id(json_data)
            result = json_data["Result"]

            if result == "DISABLED":
                logging.debug("%s: 'DISABLED' -> ok", method_id)
                return True

            if result not in result_level:
                logging.error("Unknown result '%s' for test '%s'", result, method_id)
                return False

            actual_level = result_level[result]

            if method_id in xfails:
                expected_level = xfails[method_id]
                seen_xfails.add(method_id)
                expected_label = [k for k, v in result_level.items() if v == expected_level][0]

                if actual_level != expected_level:
                    # Test is doing other than expected -> error
                    logging.error(
                        "Error: %s has result '%s' but is xfail as '%s'. "
                        "Remove or update the xfail entry.",
                        method_id, result, expected_label)
                    success = False
                else:
                    logging.debug("%s: '%s' -> ok (xfail as %s)", method_id, result, expected_label)
                    success = True
            else:
                # Not in xfail list: must strictly succeed
                if actual_level > result_level["STRICTLY_SUCCEEDED"]:
                    logging.error(failing_test_info(json_data, method_id, "STRICTLY_SUCCEEDED"))
                    success = False
                else:
                    logging.debug("%s: '%s' -> ok", method_id, result)
                    success = True

        except (KeyError, TypeError) as e:
            logging.error("Json file '%s' has missing entries: %s", result_path, e)

    return success


RESULT_FILE_NAME = "_testRun.json"


def main(args=None):
    """Parse args and check all result container files"""
    if args is None:
        args = sys.argv[1:]

    parser = argparse.ArgumentParser()
    parser.add_argument("--verbose", action="store_true", default=False)
    parser.add_argument("side", help="which side of the protocol was tested (client or server)")
    parser.add_argument("results-dir", help="directory of TLS-Anvil test results")

    args = vars(parser.parse_args(args))

    logging.basicConfig(
        level=(logging.DEBUG if args["verbose"] else logging.INFO),
        format="%(message)s",
    )

    side = args["side"]

    if side not in ["client", "server"]:
        print("Unexpected side %s" % (side))
        return 1

    results_dir = args["results-dir"]

    if not os.access(results_dir, os.X_OK):
        raise FileNotFoundError("Unable to read TLS-Anvil results dir")

    xfails = xfail_list(side)
    seen_xfails = set()

    failed_methods_count = 0
    total_methods_count = 0
    for root, _, files in os.walk(results_dir):
        for file in files:
            if file == RESULT_FILE_NAME:
                abs_path = os.path.abspath(os.path.join(root, file))
                total_methods_count += 1
                if not process_test_result(abs_path, xfails, seen_xfails):
                    failed_methods_count += 1

    if total_methods_count == 0:
        logging.error("No test results found in '%s'", results_dir)
        return 1

    # Check that every xfail entry was actually seen in the results
    missing_xfails = set(xfails.keys()) - seen_xfails
    for method_id in sorted(missing_xfails):
        expected_label = [k for k, v in result_level.items() if v == xfails[method_id]][0]
        logging.warning(
            "xfailed test '%s' (expected %s) was not found in results.",
            method_id, expected_label)

    logging.info(
        "(%i/%i) test methods successful.",
        total_methods_count - failed_methods_count,
        total_methods_count,
    )
    total_success = failed_methods_count == 0
    logging.info("Total result: %s", "Success." if total_success else "Failed.")

    return int(not total_success)


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python3

"""
Run all tests under valgrind in a thread pool

(C) 2025 Jack Lloyd

Botan is released under the Simplified BSD License (see license.txt)
"""

import multiprocessing
import optparse # pylint: disable=deprecated-module
import subprocess
import sys
import time

from multiprocessing.pool import ThreadPool

def get_concurrency():
    def_concurrency = 2
    max_concurrency = 16

    try:
        return min(max_concurrency, multiprocessing.cpu_count())
    except ImportError:
        return def_concurrency

def available_tests(botan_test):
    cmd = [botan_test, '--list-tests']
    tests = subprocess.Popen(cmd, close_fds=True, stdout=subprocess.PIPE).communicate()

    return [str(s, encoding='utf8') for s in tests[0].split()]

def run_valgrind(options, test):
    valgrind_cmd = ['valgrind', '-v', '--error-exitcode=9']

    if options.with_leak_check:
        valgrind_cmd += ['--leak-check=full', '--show-reachable=yes']

    if options.track_origins:
        valgrind_cmd += ['--track-origins=yes']

    botan_test_options = ['--test-threads=1', '--run-memory-intensive-tests']
    cmd = valgrind_cmd + [options.test_binary] + botan_test_options + test

    start = time.time()
    proc = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    duration = time.time() - start

    if options.verbose:
        print("Testing '%s' took %.02fs" % (test, duration))
        sys.stdout.flush()

    if proc.returncode == 0:
        return True # success

    print("FAILED: valgrind testing %s failed with error code %d" % (test, proc.returncode))
    print(proc.stdout.decode('utf8'))
    print(proc.stderr.decode('utf8'))
    return False

def filter_tests(available, cmdline, options):
    skip_tests = sum([x.split(',') for x in options.skip_tests], [])

    to_run = []

    for test in available:
        if test in skip_tests:
            continue

        if test.startswith('pkcs11'):
            continue

        if cmdline == [] or test in cmdline:
            to_run.append(test)

    return to_run

def split_list(list, n):
    return [list[x:x+n] for x in range(0, len(list), n)]

def main(args = None):
    if args is None:
        args = sys.argv

    parser = optparse.OptionParser()

    parser.add_option('--verbose', action='store_true', default=False, help='be noisy')

    parser.add_option('--test-binary', metavar='PATH', default='./botan-test',
                      help='path to botan-test binary')

    parser.add_option('--jobs', metavar='J', default=get_concurrency(),
                      help='number of jobs to run in parallel (default %default)')

    parser.add_option('--with-leak-check', action='store_true', default=False,
                      help='enable full valgrind leak checks')

    parser.add_option('--track-origins', action='store_true', default=False,
                      help='enable origin tracking')

    parser.add_option('--skip-tests', metavar='TESTS', default=[], action='append',
                      help='skip the named tests')

    parser.add_option('--bunch', action='store_true', default=False,
                      help='run several test suites under each valgrind exec')

    (options, args) = parser.parse_args(args)

    tests = filter_tests(available_tests(options.test_binary), args[1:], options)

    jobs = int(options.jobs)
    pool = ThreadPool(jobs)

    results = []

    if options.bunch:
        bunching = len(tests) // (jobs * 8)

        for test in split_list(tests, bunching):
            results.append(pool.apply_async(run_valgrind, (options, test)))
    else:
        for test in tests:
            results.append(pool.apply_async(run_valgrind, (options, [test])))

    fail_cnt = 0
    for result in results:
        if not result.get():
            fail_cnt += 1

    if fail_cnt > 0:
        print("%d tests failed" % (fail_cnt))
        return 1
    else:
        return 0

if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python3
# coding=utf8

"""
This script reports the sizes of various binary artifacts in CI

(C) 2025 Jack Lloyd

Botan is released under the Simplified BSD License (see license.txt)
"""

import os
import sys
import json

def format_size(bytes):
    kB = 1024

    if bytes > kB:
        return "%.02f kB" % (bytes / kB)
    else:
        return "%d bytes" % (bytes)

def report_size(fsname):
    if not os.access(fsname, os.R_OK):
        print("ERROR: Could not find %s" % (fsname))
        return
    bytes = os.stat(fsname).st_size
    print("File '%s' is %s" % (fsname, format_size(bytes)))

def main(args = None):
    if args is None:
        args = sys.argv

    if len(args) != 2:
        print("Usage: %s <build_config.json>" % (args[0]))
        return 1

    build_config = json.loads(open(args[1]).read())

    for lib in build_config['library_targets'].split(' '):
        report_size(lib)

    if 'text_exe' in build_config:
        report_size(build_config['test_exe'])

    if 'cli_exe' in build_config:
        report_size(build_config['cli_exe'])

if __name__ == '__main__':
    sys.exit(main())
